! WHIZARD 1.20 Feb 11 2002
! 
! (C) 1999-2001 by Wolfgang Kilian <kilian@particle.uni-karlsruhe.de>
!
! WHIZARD is free software; you can redistribute it and/or modify it
! under the terms of the GNU General Public License as published by 
! the Free Software Foundation; either version 2, or (at your option)
! any later version.
!
! WHIZARD is distributed in the hope that it will be useful, but
! WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
! GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this program; if not, write to the Free Software
! Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! This file has been stripped of most comments.  For documentation, refer
! to the source 'whizard.nw'

module density_vectors

  use kinds, only: default !NODEP!
  use diagnostics, only: msg_fatal
  use process_interface, only: process_scatter, process_scatter_colored !NODEP!

  implicit none
  private

  type, public :: state_density
     private
     integer :: n_states
     integer :: n_col, n_hel, n_flv
     real(kind=default), dimension(:,:,:), pointer :: p
  end type state_density

  type, public :: single_density
     integer :: n_states
     real(kind=default), dimension(:), pointer :: p
  end type single_density


  public :: create, destroy
  public :: read, write
  public :: read_dimensions_raw, write_dimensions_raw
  public :: nullify
  public :: set
  public :: uniform_density
  public :: normalize
  public :: compute_state_density
  public :: norm
  public :: select_state

  interface create
     module procedure state_density_create
     module procedure single_density_create
  end interface
  interface destroy
     module procedure state_density_destroy
     module procedure single_density_destroy
  end interface
  interface write
     module procedure state_density_write_unit
     module procedure single_density_write_unit
  end interface
  interface write
     module procedure state_density_write_raw
  end interface
  interface read
     module procedure state_density_read_raw
  end interface
  interface nullify
     module procedure state_density_nullify
  end interface
  interface set
     module procedure state_density_set_entry
  end interface
  interface uniform_density
     module procedure uniform_single_density
  end interface
  interface normalize
     module procedure state_density_normalize
     module procedure single_density_normalize
  end interface
  interface norm
     module procedure state_density_norm
     module procedure single_density_norm
  end interface

contains

  subroutine state_density_create (rho, n_col, n_hel, n_flv)
    type(state_density), intent(inout) :: rho
    integer, intent(in), optional :: n_col, n_hel, n_flv
    if (present(n_col))  rho%n_col = n_col
    if (present(n_hel))  rho%n_hel = n_hel
    if (present(n_flv))  rho%n_flv = n_flv
    if (rho%n_col==0)  rho%n_col = 1
    if (rho%n_hel==0)  rho%n_hel = 1
    if (rho%n_flv==0)  rho%n_flv = 1
    rho%n_states = rho%n_col * rho%n_hel * rho%n_flv
    allocate (rho%p (rho%n_col, rho%n_hel, rho%n_flv))
    rho%p = 0
  end subroutine state_density_create

  subroutine single_density_create (rho, n_states)
    type(single_density), intent(inout) :: rho
    integer, intent(in) :: n_states
    rho%n_states = n_states;   if (rho%n_states==0)  rho%n_states = 1
    allocate (rho%p(rho%n_states))
    rho%p = 0
  end subroutine single_density_create

  subroutine state_density_destroy (rho)
    type(state_density), intent(inout) :: rho
    deallocate (rho%p)
    rho%n_col = 0
    rho%n_hel = 0
    rho%n_flv = 0
    rho%n_states = 0
  end subroutine state_density_destroy

  subroutine single_density_destroy (rho)
    type(single_density), intent(inout) :: rho
    deallocate (rho%p)
    rho%n_states = 0
  end subroutine single_density_destroy

  subroutine state_density_write_unit (u, rho)
    integer, intent(in) :: u
    type(state_density), intent(in) :: rho
    integer :: f, h
    write (u,*) "State density vector:"
    do f=1, ubound(rho%p, dim=3)
       write (u,*)  "Flavor:", f
       do h=1, ubound(rho%p, dim=2)
          write (u,*)  "Helicity:", h
          write (u,*)  rho%p(:,h,f)
       end do
    end do
  end subroutine state_density_write_unit

  subroutine single_density_write_unit (u, rho)
    integer, intent(in) :: u
    type(single_density), intent(in) :: rho
    write (u,*)  rho%p
  end subroutine single_density_write_unit

  subroutine state_density_write_raw (u, rho, raw)
    integer, intent(in) :: u
    type(state_density), intent(in) :: rho
    logical, intent(in) :: raw
    if (raw) then
       write (u) rho%p
    end if
  end subroutine state_density_write_raw

  subroutine state_density_read_raw (u, rho, raw)
    integer, intent(in) :: u
    type(state_density), intent(inout) :: rho
    logical, intent(in) :: raw
    integer :: n_col, n_hel, n_flv
    if (raw) then
       read (u) rho%p
    end if
  end subroutine state_density_read_raw

  subroutine write_dimensions_raw (u, rho)
    integer, intent(in) :: u
    type(state_density), intent(in) :: rho
    write (u) rho%n_col, rho%n_hel, rho%n_flv
  end subroutine write_dimensions_raw

  subroutine read_dimensions_raw (u, rho)
    integer, intent(in) :: u
    type(state_density), intent(inout) :: rho
    read (u) rho%n_col, rho%n_hel, rho%n_flv
  end subroutine read_dimensions_raw

  subroutine state_density_nullify (rho)
    type(state_density), intent(inout) :: rho
    rho%p = 0
  end subroutine state_density_nullify

  subroutine state_density_set_entry (rho, col, hel, flv, val)
    type(state_density), intent(inout) :: rho
    integer, intent(in) :: col, hel, flv
    real(kind=default), intent(in) :: val
    rho%p(col,hel,flv) = val
  end subroutine state_density_set_entry

  subroutine uniform_single_density (rho)
    type(single_density), intent(inout) :: rho
    rho%p = 1._default / rho%n_states
  end subroutine uniform_single_density

  subroutine state_density_normalize (rho)
    type(state_density), intent(inout) :: rho
    if (any(rho%p/=0))  rho%p = rho%p / sum(rho%p)
  end subroutine state_density_normalize

  subroutine single_density_normalize (rho)
    type(single_density), intent(inout) :: rho
    if (any(rho%p/=0))  rho%p = rho%p / sum(rho%p)
  end subroutine single_density_normalize

  subroutine compute_state_density &
       & (p, rho_in, rho_out, process_id, colored, ok)
    real(kind=default), dimension(0:,:), intent(in) :: p
    type(state_density), intent(in) :: rho_in
    type(state_density), intent(inout) :: rho_out
    character(len=*), intent(in) :: process_id
    logical, intent(in) :: colored
    logical, intent(out) :: ok
    if (colored) then
       call process_scatter_colored (p, rho_in%p(1,:,:), rho_out%p, process_id)
    else
       call process_scatter (p, rho_in%p(1,:,:), rho_out%p(1,:,:), process_id)
    end if
    ok = all (rho_out%p >= 0)
  end subroutine compute_state_density

  function state_density_norm (rho) result (r)
    type(state_density), intent(in) :: rho
    real(kind=default) :: r
    r = sum(rho%p)
  end function state_density_norm

  function single_density_norm (rho) result (r)
    type(single_density), intent(in) :: rho
    real(kind=default) :: r
    r = sum(rho%p)
  end function single_density_norm

  subroutine select_state (rho, r, col, hel, flv)
    type(state_density), intent(in) :: rho
    real(kind=default), intent(in) :: r
    integer, intent(out) :: col, hel, flv
    real(kind=default) :: rho_sum, rho_comp
    rho_comp = r * norm (rho)
    rho_sum = 0
    LOOP_FLV: do flv=1, rho%n_flv
       LOOP_HEL: do hel=1, rho%n_hel
          LOOP_COL: do col=1, rho%n_col
             rho_sum = rho_sum + rho%p(col,hel,flv)
             if (rho_comp < rho_sum)  exit LOOP_FLV
          end do LOOP_COL
       end do LOOP_HEL
    end do LOOP_FLV
  end subroutine select_state


end module density_vectors
