! WHIZARD 1.20 Feb 11 2002
! 
! (C) 1999-2001 by Wolfgang Kilian <kilian@particle.uni-karlsruhe.de>
!
! WHIZARD is free software; you can redistribute it and/or modify it
! under the terms of the GNU General Public License as published by 
! the Free Software Foundation; either version 2, or (at your option)
! any later version.
!
! WHIZARD is distributed in the hope that it will be useful, but
! WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
! GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this program; if not, write to the Free Software
! Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! This file has been stripped of most comments.  For documentation, refer
! to the source 'whizard.nw'

module whizard

  use kinds, only: default !NODEP!
  use kinds, only: double !NODEP!
  use vamp_interface, only: mc_integrate, mc_enable_grid, mc_generate
  use vamp_interface, only: mc_recalculate
  use analyze, only: analysis_bundle, create, destroy, apply, read, write
  use analyze, only: TEXT_FMT, GML_FMT
  use monte_carlo_states, only: mcs, monte_carlo_state
  use monte_carlo_states, only: mc_process_index, mc_number_processes
  use monte_carlo_states, only: get_number_processes
  use monte_carlo_states, only: current_event, rng
  use monte_carlo_states, only: monte_carlo_input, mpi90_broadcast
  use monte_carlo_states, only: create, destroy, read, write
  use monte_carlo_states, only: mc_write_header, mc_write_hline, mc_write_dline
  use monte_carlo_states, only: mc_write_result
  use monte_carlo_states, only: checksum, checksum_error
  use monte_carlo_states, only: JETSET_FRAGMENTATION, PYTHIA_FRAGMENTATION
  use hepevt_common, only: hepevt_fill, hepevt_write
  use beams, only: beams_pythia_format
  use events, only: read, write, event, destroy, destroy_rho_in
  use events, only: RAW_FMT, JETSET_FMT1, JETSET_FMT2, JETSET_FMT3, STDHEP_FMT
  use density_vectors, only: create, destroy
  use particles, only: particle_mass, particle_momentum, particle_code
  use particles, only: particle_color
  use lorentz, only: array
  use histograms, only: histogram, create_histogram !NODEP!
  use histograms, only: fill_histogram, write_histogram !NODEP!
  use diagnostics, only: msg_buffer, msg_count_clear, msg_summary
  use diagnostics, only: msg_message, msg_warning, msg_error, msg_fatal
  use diagnostics, only: msg_terminate
  use mpi90, only: mpi90_rank !NODEP!
  use mpi90, only: mpi90_init, mpi90_broadcast, mpi90_finalize !NODEP!
  use file_utils, only: free_unit !NODEP!
  use limits, only: DEFAULT_FILENAME, FILENAME_LEN, VERSION_STRLEN
  use limits, only: PYTHIA_BEAM_STRLEN
  use limits, only: WHIZARD_ROOT
  use limits, only: WEIGHT_HISTOGRAMS_NBIN
  use process_interface, only: processes_list !NODEP!
  use stdhep_interface, only: stdhep_init, stdhep_end !NODEP!
  use jetset_interface, only: jetset_suppress_banner !NODEP!
  use jetset_interface, only: jetset_open_write, jetset_close !NODEP!
  use jetset_interface, only: pythia_init, jetset_init !NODEP!
  use jetset_interface, only: jetset_fragment, pythia_fragment !NODEP!
  use jetset_interface, only: pythia_print_statistics !NODEP!
  use jetset_interface, only: new_event_needed_pythia !NODEP!
  use jetset_interface, only: number_events_whizard !NODEP!
  use jetset_interface, only: number_events_pythia !NODEP!
  use tao_random_numbers, only: tao_random_number !NODEP!

  implicit none
  private

  public :: whizard_set_filename
  public :: whizard_read_input
  public :: whizard_write_output
  public :: whizard_integrate
  public :: whizard_number_processes
  public :: whizard_get_results
  public :: whizard_generate
  public :: whizard_events_begin
  public :: whizard_event
  public :: whizard_events_end
  public :: whizard_end

  public :: whizard_input

  type(monte_carlo_input) :: whizard_input
  character(len=FILENAME_LEN) :: whizard_filename = DEFAULT_FILENAME
  type(analysis_bundle), dimension(:), allocatable :: whizard_analysis
  logical :: whizard_input_exists = .false.
  logical :: mcs_exists = .false.
  logical :: event_files_open = .false.
  logical :: mpi_initialized = .false.
  integer :: proc_id
  integer :: uraw, uevt, uwgt
  real(kind=default) :: total_integral, total_error, total_efficiency
  real(kind=default), dimension(:), allocatable :: process_probability


  save

contains

  subroutine whizard_set_filename (filename)
    character(len=*), intent(in) :: filename
    if (len(filename) > FILENAME_LEN) then
       call msg_warning (" Filename too long; will be truncated.")
    end if
    whizard_filename = trim(filename)
  end subroutine whizard_set_filename

  subroutine whizard_read_input
    integer :: n_channels
    if (whizard_input_exists .or. mcs_exists)  call whizard_end
    if (.not.mpi_initialized)  then
       call mpi90_init;  mpi_initialized = .true.
    end if
    call mpi90_rank (proc_id)
    call create (whizard_input)
    call msg_message (" WHIZARD 1.20 (Feb 11 2002)")
    if (proc_id == WHIZARD_ROOT) then
       call read (trim(whizard_filename), whizard_input, n_channels)
    end if
    call mpi90_broadcast (whizard_input, WHIZARD_ROOT)
    call write_global_log_file
    whizard_input_exists = .true.
  contains
    subroutine write_global_log_file
      integer :: u
      u = free_unit ()
      open (u, file=trim(whizard_filename)//".out", action="write")
      call msg_message (" WHIZARD 1.20 (Feb 11 2002)", unit=u)
      call processes_list (u)
      call msg_message &
           & (" For the process-specific logs, consult the files " &
           &  // trim(whizard_filename) //".XXX.out", unit=u)
      call msg_message (" Input data, including default values:", unit=u)
      call write (u, whizard_input, n_channels)
      close (u)
      call msg_message (" Wrote " // trim(whizard_filename) // ".out")
    end subroutine write_global_log_file
  end subroutine whizard_read_input

  subroutine whizard_write_output
    call mpi90_rank (proc_id)
    if (proc_id == WHIZARD_ROOT) then
       if (whizard_input_exists) then
          if (mcs_exists) then
             call write (trim(whizard_filename), mcs, whizard_input)
          else
             mc_number_processes = &
                  & get_number_processes (whizard_input%process_id)
             allocate (mcs(mc_number_processes))
             call create (mcs, rng, current_event, &
                  &       whizard_input, trim(whizard_filename))
             call write (trim(whizard_filename), mcs, whizard_input)
             call destroy (mcs)
             deallocate (mcs)
             call destroy (current_event)
          end if
       else
          call msg_warning &
               & (" Output file requested before input is known.  Ignored.")
       end if
    end if
  end subroutine whizard_write_output

  subroutine whizard_integrate
    real(kind=default) :: time_per_event
    real(kind=default), dimension(:), allocatable :: tmp
    if (mcs_exists)  call whizard_end
    if (.not.whizard_input_exists)  call whizard_read_input
    call mpi90_rank (proc_id)
    mc_number_processes = &
         & get_number_processes (whizard_input%process_id)
    if (mc_number_processes == 0) then
       call processes_list (6)
       call msg_terminate (" No process selected in input file: exiting.")
    end if
    allocate (mcs(mc_number_processes))
    call create &
         & (mcs, rng, current_event, whizard_input, trim(whizard_filename))
    mcs_exists = .true.
    do mc_process_index=1, mc_number_processes
       call mc_integrate (mcs(mc_process_index), &
            &             whizard_input, filename=trim(whizard_filename), &
            &             time_per_event=time_per_event)
       call show_time_estimate
    end do
    total_integral = sum (mcs%integral)
    total_error = sqrt (sum (mcs%error**2))
    allocate (tmp(mc_number_processes))
    where (mcs%efficiency /= 0)
       tmp = mcs%integral / mcs%efficiency
    elsewhere
       tmp = 0
    end where
    if (sum(tmp) /= 0) then
       total_efficiency = total_integral / sum(tmp)
    else
       total_efficiency = 0
    end if
    deallocate (tmp)
    allocate (process_probability(mc_number_processes))
    if (total_integral /= 0) then
       process_probability = mcs%integral / total_integral
    else
       process_probability = 1 / real(mc_number_processes, kind=default)
    end if
    call mc_write_dline (6)
    call msg_message (" Total cross section summary (all processes):")
    call mc_write_header (6)
    call mc_write_result &
         & (6, 0, 0, total_integral, total_error, total_efficiency, 0._default)
    call mc_write_hline (6)
    call msg_message (" Integration complete.")
    if (proc_id == WHIZARD_ROOT) then
       if (mcs(1)%screen_diagnostics) then
          do mc_process_index=1, mc_number_processes
             call write (6, mcs(mc_process_index), whizard_input)
          end do
       end if
    end if
  contains
    subroutine show_time_estimate
      if (time_per_event > 0) then
         call msg_message
         write (msg_buffer, "(1x,A,I9,A,I2.2,A,I2.2,A)") &
             & "Time estimate for generating 10000 unweighted events:", &
             & int(10000*time_per_event)/3600, "h ", &
             & mod(int(10000*time_per_event), 3600)/60, "m ", &
             & mod(int(10000*time_per_event), 60), "s"
         call msg_message
      end if
    end subroutine show_time_estimate
  end subroutine whizard_integrate

  subroutine whizard_number_processes (n)
    integer, intent(out) :: n
    n = 0
    if (mcs_exists) then
       n = mc_number_processes
    else
       call msg_warning (" Number of processes requested but not yet known.")
    end if
  end subroutine whizard_number_processes

  subroutine whizard_get_results &
       & (process_index, integral, error, chi2, efficiency)
    integer, intent(in) :: process_index
    real(kind=double), intent(out) :: integral, error, chi2, efficiency
    integral = 0
    error = 0
    chi2 = 0
    efficiency = 0
    if (mcs_exists) then
       if (process_index == 0) then
          integral = total_integral
          error = total_error
          chi2 = 0
          efficiency = total_efficiency
       else if (process_index<0 .or. process_index > mc_number_processes) then
          call msg_error &
               & (" Integration results for illegal process index requested")
       else
          integral = mcs(process_index)%integral
          error = mcs(process_index)%error
          chi2 = mcs(process_index)%chi2
          efficiency = mcs(process_index)%efficiency
       end if
    else
       call msg_warning (" Integration results requested but not yet known.")
    end if
  end subroutine whizard_get_results

  subroutine whizard_generate
    integer :: ievt, n_calls
    real(kind=default) :: lumi
    call get_n_calls (mcs(1), n_calls, lumi)
    mcs%luminosity = lumi
    if (n_calls > 0) then
       call whizard_events_begin (n_calls)
       if (mcs(1)%unweighted) then
          write (msg_buffer, "(1x,A,1x,G11.4)") &
                  & "Event sample corresponds to luminosity [fb-1] =", lumi
          call msg_message
          if (mcs(1)%read_events_raw) then
             write (msg_buffer, "(1x,A,1x,I10,1x,A)") &
                  & "Reading", n_calls, "unweighted events ..."
          else
             write (msg_buffer, "(1x,A,1x,I10,1x,A)") &
                  & "Generating", n_calls, "unweighted events ..."
          end if
          call msg_message
          if (mcs(1)%recalculate) then
             call msg_message (" (recalculating matrix element values)")
          end if
          do ievt=1, n_calls
             call whizard_event
          end do
       else
          if (mcs(1)%read_events_raw) then
             write (msg_buffer, "(1x,A,1x,I10,1x,A)") &
                  & "Reading", n_calls, "weighted events per process ..."
          else
             write (msg_buffer, "(1x,A,1x,I10,1x,A)") &
                  & "Generating", n_calls, "weighted events per process ..."
          end if
          call msg_message
          call generate_events_weighted (n_calls)
       end if
       call whizard_events_end
    else
       call msg_message (" No event generation requested")
    end if
  contains
    subroutine get_n_calls (mcs, n_calls, lumi)
      type(monte_carlo_state), intent(in) :: mcs
      integer, intent(out) :: n_calls
      real(kind=default), intent(out) :: lumi
      if (total_integral > 0) then
         if (mcs%unweighted) then
            lumi = max (mcs%luminosity, mcs%calls(2,5) / total_integral)
            n_calls = mcs%calls(1,5) * nint (total_integral * lumi)
         else
            if (mcs%luminosity > 0) then
               call msg_warning (" Luminosity ignored for weighted events")
               lumi = 0
            end if
            n_calls = mcs%calls(1,5) * mcs%calls(2,5)
         end if
      else
         lumi = 0
         n_calls = 0
      end if
    end subroutine get_n_calls
  end subroutine whizard_generate

  subroutine whizard_events_begin (n_calls)
    integer, intent(in) :: n_calls
    logical :: pythia_first, jetset_first
    if (event_files_open)  call whizard_events_end
    if (.not.mcs_exists)  call whizard_integrate
    call mpi90_rank (proc_id)
    if (total_integral <= 0)  &
         & call msg_fatal ("Integral must be positive for simulation.")
    jetset_first = .true.
    pythia_first = .true.
    allocate (whizard_analysis(mc_number_processes))
    do mc_process_index=1, mc_number_processes
       call mc_enable_grid (mcs(mc_process_index), &
            &               mcs(mc_process_index)%g_best, &
            &               mcs(mc_process_index)%safety_factor)
       if (mcs(mc_process_index)%fragment) &
            & call prepare_fragmentation &
            &   (mcs(mc_process_index), maxval(mcs%n_tot), &
            &    jetset_first, pythia_first)
       call create (whizard_analysis(mc_process_index))
       if (proc_id == WHIZARD_ROOT) then
          call read (trim(whizard_filename), &
               &     whizard_analysis(mc_process_index), &
               &     trim(mcs(mc_process_index)%process_id_current))
       end if
    end do
!    call mpi90_broadcast (whizard_analysis, WHIZARD_ROOT)
    call open_raw_events_file (n_calls)
    call open_output_events_file (n_calls)
    current_event%count = 0
    event_files_open = .true.
  contains
    subroutine open_raw_events_file (n_calls)
      integer, intent(in) :: n_calls
      integer :: chsum1, chsum2
      character(len=VERSION_STRLEN) :: version_string
      if (mcs(1)%read_events_raw) then
         if (proc_id == WHIZARD_ROOT) then
            call msg_message (" Looking for raw event file "// &
                 &            trim(whizard_filename)//".evx ...")
            inquire(file=trim(whizard_filename)//".evx", &
                 &  exist=mcs(1)%read_events_raw)
            if (.not.mcs(1)%read_events_raw) &
                 & call msg_message (" ... not found.  Will be generated.")
         end if
         call mpi90_broadcast (mcs(1)%read_events_raw, WHIZARD_ROOT)
      end if
      chsum1 = checksum (mcs(1),5)
      if (.not.mcs(1)%unweighted) then
         chsum1 = chsum1 + 13579*n_calls
      end if
      if (mcs(1)%read_events_raw) then
         if (proc_id == WHIZARD_ROOT) then
            uraw = free_unit()
            open (unit=uraw, file=trim(whizard_filename)//".evx", &
                 & status='old', action='read', form='unformatted')
            read(uraw) chsum2
            read(uraw) version_string
            if (version_string /= "WHIZARD version 1.20 (Feb 11 2002)") then
               call msg_warning (" Raw event file has been generated by " // &
                    &            trim(version_string))
            end if
         end if
      else if (mcs(1)%write_events_raw) then
         if (proc_id == WHIZARD_ROOT) then
            uraw = free_unit()
            open(unit=uraw, file=trim(whizard_filename)//".evx", &
                 & status='replace', action='write', form='unformatted')
            chsum2 = chsum1
            write(uraw) chsum2
            version_string = "WHIZARD version 1.20 (Feb 11 2002)"
            write(uraw) version_string
         end if
      else
         chsum2 = chsum1
      end if
      call mpi90_broadcast (chsum1, WHIZARD_ROOT)
      call mpi90_broadcast (chsum2, WHIZARD_ROOT)
      if (.not.mcs(1)%read_events_force .and. chsum1 /= chsum2) &
           & call checksum_error
    end subroutine open_raw_events_file
    subroutine open_output_events_file (n_calls)
      integer, intent(in) :: n_calls
      if (proc_id == WHIZARD_ROOT .and. mcs(1)%write_events) then
         if (mcs(1)%unweighted) then
            select case (mcs(1)%write_events_format)
            case (JETSET_FMT1, JETSET_FMT2, JETSET_FMT3)
               call msg_message (" Writing events to file "// &
                    &            trim(whizard_filename)//".pythia")
               uevt = free_unit ()
               call jetset_open_write &
                    & (unit=uevt, file=trim(whizard_filename)//".pythia")
            case (STDHEP_FMT)
               call msg_message (" Writing events to file "// &
                    &            trim(whizard_filename)//".stdhep")
               call stdhep_init (trim(whizard_filename)//".stdhep", &
                    &            "WHIZARD event sample", n_calls)
            case default
               call msg_message (" Writing events to file "// &
                    &            trim(whizard_filename)//".evt")
               uevt = free_unit()
               open(unit=uevt, file=trim(whizard_filename)//".evt", &
                    & status='replace', action='write')
            end select
         else
            call msg_error (" Writing weighted events to file is not supported")
         end if
         if (mcs(1)%screen_events) then
            select case (mcs(1)%write_events_format)
            case (STDHEP_FMT)
               call msg_error (" Can't display events in STDHEP format")
            case default
               call msg_message (" Displaying generated events ...")
            end select
         end if
      end if
    end subroutine open_output_events_file
    subroutine prepare_fragmentation &
         & (mcs, n_tot_max, jetset_first, pythia_first)
      type(monte_carlo_state), intent(inout) :: mcs
      integer, intent(in) :: n_tot_max
      logical, intent(inout) :: jetset_first, pythia_first
      character(len=PYTHIA_BEAM_STRLEN) :: beam_str, target_str
      select case (mcs%fragmentation_method)
      case (JETSET_FRAGMENTATION)
         if (.not.mcs%show_pythia_banner)  call jetset_suppress_banner
         if (jetset_first) then
            call jetset_init (mcs%pythia_parameters)
            jetset_first = .false.
         end if
         if (.not.color_flow_known (mcs%code(:,1), mcs%n_col)) then
            call msg_warning &
                 & (" The color flow in process " &
                 &  //trim(mcs%process_id_current)//" is unknown.")
            call msg_message &
                 & (" JETSET will deduce the color configuration " &
                 &     //"from the particle ordering.")
         else if (.not.color_flow_known (mcs%code_in(:,1), 0)) then
            call msg_warning &
                 & (" The color flow in process " &
                 &  //trim(mcs%process_id_current) &
                 &  //" involves the initial state.")
            call msg_message &
                 & (" JETSET fragmentation treats the final state only.")
            call msg_message &
                 & (" To keep all color connections, " &
                 &  //"use PYTHIA fragmentation instead.")
         end if
      case (PYTHIA_FRAGMENTATION)
         if (.not.mcs%show_pythia_banner)  call jetset_suppress_banner
         if (color_flow_known (mcs%code(:,1), mcs%n_col)) then
            if (pythia_first) then
               call beams_pythia_format &
                    & (mcs%beam, mcs%code_in(:,1), beam_str, target_str)
               call pythia_init (n_tot_max, mcs%process_id, &
                    &            trim(beam_str), trim(target_str), &
                    &            all(mcs%beam%fixed_energy), &
                    &            mcs%sqrts, total_integral, &
                    &            mcs%pythia_processes, &
                    &            mcs%pythia_parameters, &
                    &            mcs%show_pythia_initialization)
               jetset_first = .false.
               pythia_first = .false.
            end if
         else
            call msg_error &
                 & (" PYTHIA fragmentation is unavailable for this process.")
            call msg_message &
                 & (" Reason: The color flow in process " &
                 &  //trim(mcs%process_id)//" is unknown.")
            call msg_message &
                 & (" I will fall back to JETSET fragmentation.")
            call msg_message &
                 & (" JETSET will deduce the color configuration " &
                 &     //"from the particle ordering.")
            mcs%fragmentation_method = JETSET_FRAGMENTATION
            if (jetset_first) then
               call jetset_init (mcs%pythia_parameters)
               jetset_first = .false.
            end if
         end if
      case default
         call jetset_suppress_banner
      end select
    end subroutine prepare_fragmentation

    function color_flow_known (code, n_col) result (ok)
      integer, dimension(:), intent(in) :: code
      integer, intent(in) :: n_col
      logical :: ok
      integer :: i
      if (n_col > 0) then
         ok = .true.
      else
         ok = .true.
         do i=1, size(code)
            if (particle_color (code(i)) /= 0)  ok = .false.
         end do
      end if
    end function color_flow_known

  end subroutine whizard_events_begin

  subroutine whizard_event
    real(kind=default) :: weight, excess, r
    logical :: rho_in_allocated
    integer :: iostat
    rho_in_allocated = .false.
    call mpi90_rank (proc_id)
    if (new_event_needed_pythia) then
       if (.not.event_files_open) then
          call msg_warning ("The call to whizard_events_begin is missing.")
          call msg_message &
               & ("I will call whizard_events_begin with argument 0.")
          call whizard_events_begin (0)
       end if
       excess = 0
       if (mcs(1)%read_events_raw) then
          if (proc_id == WHIZARD_ROOT) then
             call read (uraw, current_event, weight=weight, &
                  &     fmt=RAW_FMT, iostat=iostat)
          end if
!           call mpi90_broadcast (current_event, WHIZARD_ROOT)
          rho_in_allocated = .true.
          call mpi90_broadcast (weight, WHIZARD_ROOT)
          call mpi90_broadcast (iostat, WHIZARD_ROOT)
          if (weight > 1._default)  excess = weight - 1._default
          if (iostat == 0) then
             current_event%count = current_event%count + 1
             mc_process_index = current_event%process_index
          else if (iostat < 0) then
             rho_in_allocated = .false.
             call reopen_raw_events_file
          else
             call msg_fatal &
                  & (" I/O error occured while reading raw event file.")
          end if
          if (mcs(1)%recalculate) then
             call mc_recalculate (mcs(mc_process_index), current_event)
          end if
       end if
       if (.not.mcs(1)%read_events_raw) then
          call tao_random_number (rng, r)
          call process_select (mc_process_index, r)
          call mc_generate (mcs(mc_process_index), excess=excess)
          if (mcs(1)%write_events_raw) then
             if (proc_id == WHIZARD_ROOT) call write &
                  & (uraw, current_event, weight=1._default+excess, &
                  &  fmt=RAW_FMT)
          end if
       end if
       if (rho_in_allocated) then
          call destroy_rho_in (current_event)
          rho_in_allocated = .false.
       end if
       call hepevt_fill (current_event)
       call apply (whizard_analysis(mc_process_index), &
            &      current_event, excess=excess)
    end if
    call fragment_current_event (mcs(mc_process_index))
  contains
    subroutine process_select (n, r)
      integer, intent(out) :: n
      real(kind=default), intent(in) :: r
      real(kind=default) :: psum
      integer :: i
      psum = 0
      n = mc_number_processes
      LOOP: do i=1, mc_number_processes
         psum = psum + process_probability(i)
         if (r < psum) then
            n = i
            exit LOOP
         end if
      end do LOOP
    end subroutine process_select
      
    subroutine reopen_raw_events_file
      mcs%read_events_raw = .false.
      if (proc_id == WHIZARD_ROOT) then
         close (uraw)
         write (msg_buffer, "(1x,A,1x,I10,1x,A)") &
              & " End of file reached after reading", &
              & current_event%count, "events"  
         call msg_warning
         if (mcs(1)%recalculate) then
            call msg_message (" Matrix element recalculation required, but no reference values available.")
            call msg_fatal (" Event generation must be stopped.")
         else
            call msg_message (" Generating fresh events ...")
            if (mcs(1)%write_events_raw) then
               open(unit=uraw, file=trim(whizard_filename)//".evx", &
                    & status='old', position='append', &
                    & action='write', form='unformatted')
            end if
         end if
      end if
    end subroutine reopen_raw_events_file
    subroutine fragment_current_event (mcs)
      type(monte_carlo_state), intent(in) :: mcs
      if (mcs%fragment) then
         select case (mcs%fragmentation_method)
         case (JETSET_FRAGMENTATION)
            call jetset_fragment &
                 & (current_event%color_flow, current_event%anticolor_flow)
         case (PYTHIA_FRAGMENTATION)
            call pythia_fragment_call (mcs%n_tot, current_event)
         end select
         if (proc_id == WHIZARD_ROOT .and. mcs%write_events) &
              & call hepevt_write (uevt, fmt=mcs%write_events_format)
         if (proc_id == WHIZARD_ROOT .and. mcs%screen_events &
              .and. mcs%write_events_format /= STDHEP_FMT) &
              & call hepevt_write (6, fmt=mcs%write_events_format)
      else
         if (proc_id == WHIZARD_ROOT .and. mcs%write_events) & 
              & call write (uevt, current_event, &
              &             weight=1._default, fmt=mcs%write_events_format)
         if (proc_id == WHIZARD_ROOT .and. mcs%screen_events &
              & .and. mcs%write_events_format /= STDHEP_FMT) & 
              & call write (6, current_event, &
              &             weight=1._default, fmt=mcs%write_events_format)
      end if
    end subroutine fragment_current_event

    subroutine pythia_fragment_call (n_tot, evt)
      integer, intent(in) :: n_tot
      type(event), intent(in) :: evt
      integer, dimension(n_tot) :: code
      real(kind=default), dimension(n_tot, 0:3) :: p
      real(kind=default), dimension(n_tot) :: m
      integer :: i
      do i=1, 2
         code(i) = particle_code (evt%prt(-i))
         m(i) = particle_mass (evt%prt(-i))
         p(i,0:3) = array (particle_momentum (evt%prt(-i)))
      end do
      do i=3, n_tot
         code(i) = particle_code (evt%prt(i-2))
         m(i) = particle_mass (evt%prt(i-2))
         p(i,0:3) = array (particle_momentum (evt%prt(i-2)))
      end do
      call pythia_fragment (m, p, code, evt%color_flow, evt%anticolor_flow)
    end subroutine pythia_fragment_call

  end subroutine whizard_event

  subroutine whizard_events_end
    if (.not.event_files_open)  return
    mcs%it = mcs%it + 1
    if (any(mcs%fragment &
         & .and. mcs%fragmentation_method==PYTHIA_FRAGMENTATION)) then
       if (any(mcs%show_pythia_statistics))  call pythia_print_statistics
       call mc_write_dline (6)
       call msg_message (" WHIZARD/PYTHIA fragmentation summary:")
       write (msg_buffer, "(3x,A,1x,I12)") &
            & "Events generated and fragmented by PYTHIA =           ", &
            & number_events_pythia
       call msg_message
       write (msg_buffer, "(3x,A,1x,I12)") &
            & "Events generated by WHIZARD and fragmented by PYTHIA =", &
            & number_events_whizard
       call msg_message
    end if
    call msg_message (" Event generation complete.")
    call mpi90_rank (proc_id)
    if (proc_id == WHIZARD_ROOT) then
       do mc_process_index=1, mc_number_processes
          call write (6, whizard_analysis(mc_process_index), &
               &      mcs(mc_process_index), &
               &      show_histograms=mcs(mc_process_index)%screen_histograms)
       end do
       call write (trim(whizard_filename), TEXT_FMT, &
            &      whizard_analysis, mcs)
       call write (trim(whizard_filename), GML_FMT, &
            &      whizard_analysis, mcs)
    end if
    do mc_process_index=1, mc_number_processes
       call destroy (whizard_analysis(mc_process_index))
    end do
    deallocate (whizard_analysis)
    call close_raw_events_file
    call close_output_events_file
    event_files_open = .false.
  contains
    subroutine close_raw_events_file
      if (proc_id == WHIZARD_ROOT) then
         if (mcs(1)%read_events_raw .or. mcs(1)%write_events_raw)  close (uraw)
      end if
    end subroutine close_raw_events_file
    subroutine close_output_events_file
      if (proc_id == WHIZARD_ROOT .and. mcs(1)%write_events) then
         select case (mcs(1)%write_events_format)
         case (JETSET_FMT1, JETSET_FMT2, JETSET_FMT3)
            call jetset_close (uevt)
         case (STDHEP_FMT)
            call stdhep_end
         case default
            close (uevt)
         end select
      end if
    end subroutine close_output_events_file
  end subroutine whizard_events_end

  subroutine generate_events_weighted (n_calls)
    integer, intent(in) :: n_calls
    real(kind=default) :: weight, efftmp
    real(kind=default), dimension(:,:), allocatable :: eff, effsum
    type(histogram), dimension(:,:), allocatable :: hh
    type(histogram), dimension(:), allocatable :: h
    uwgt = free_unit ()
    allocate (hh(maxval(mcs%f%n_channels), mc_number_processes))
    allocate (h(mc_number_processes))
    do mc_process_index=1, mc_number_processes
       call create_histogram (h(mc_process_index), &
            & 0._default, 1._default, WEIGHT_HISTOGRAMS_NBIN)
    end do
    allocate (eff(maxval(mcs%f%n_channels), mc_number_processes))
    allocate (effsum(maxval(mcs%f%n_channels), mc_number_processes))
    eff = 0
    effsum = 0
    do mc_process_index=1, mc_number_processes
       call msg_message (" Generating events for process " &
            & // trim(mcs(mc_process_index)%process_id_current))
       open (unit=uwgt, file=trim(whizard_filename)//"." &
            &  //trim(mcs(mc_process_index)%process_id_current)//".wgt", &
            & status='replace', action='write')
       if (mcs(1)%recalculate) then
          call msg_message (" (recalculating matrix element values)")
       end if
       call generate_weighted_single (mcs(mc_process_index), &
            & h(mc_process_index), hh(:,mc_process_index), &
            & eff(:,mc_process_index), effsum(:,mc_process_index))
       if (proc_id == WHIZARD_ROOT) then
          call msg_message (unit=uwgt)
          call msg_message (" Multi-channel:", unit=uwgt)
          call write_histogram (h(mc_process_index), uwgt, over=.true.)
          call msg_message (unit=uwgt)
          write (msg_buffer, "(1x,A,1x,I12)") &
               & "Warmup events    =", &
               & mcs(mc_process_index)%calls(1,4) &
               & * mcs(mc_process_index)%calls(2,4)
          call msg_message (unit=uwgt)
          write (msg_buffer, "(1x,A,1x,I12)") &
               & "Generated events =", n_calls
          write (msg_buffer, "(1x,A,1x,F7.3)") &
               & "Efficiency (%)   =", &
               & 100 * sum(effsum(:,mc_process_index)) / n_calls
       end if
       close (uwgt)
    end do
    deallocate (h, hh)
    deallocate (eff, effsum)

  contains

    subroutine generate_weighted_single (mcs, h, hh, eff, effsum)
      type(monte_carlo_state), intent(inout) :: mcs
      type(histogram), intent(inout) :: h
      type(histogram), dimension(:), intent(inout) :: hh
      real(kind=default), dimension(:), intent(inout) :: eff, effsum
      integer :: ievt, ch, iostat
      logical :: rho_in_allocated
      call msg_message (" WHIZARD run for process " &
           & // trim(mcs%process_id_current), unit=uwgt)
      call msg_message (" Weight distribution:", unit=uwgt)
      rho_in_allocated = .false.
      do ch=1, mcs%f%n_channels
         if (mcs%g%num_calls(ch) < 2) cycle
         call msg_message (unit=uwgt)
         write (msg_buffer, "(1x,A,1x,I4)") "Channel #", ch
         call msg_message (unit=uwgt)
         call create_histogram (hh(ch), &
              & 0._default, 1._default, WEIGHT_HISTOGRAMS_NBIN)
         do ievt=1, int(n_calls * mcs%g%weights(ch))
            if (mcs%read_events_raw) then
               current_event%count = ievt
               if (proc_id == WHIZARD_ROOT) &
                    & call read(uraw, current_event, fmt=RAW_FMT, &
                    &           probability=eff(ch), &
                    &           weight=weight, channel=ch, &
                    &           iostat=iostat)
!             call mpi90_broadcast (current_event, WHIZARD_ROOT)
               rho_in_allocated = .true.
               if (iostat == 0) then
                  current_event%count = current_event%count + 1
               else if (iostat < 0) then
                  call msg_fatal ("Unexpected end of raw event file.")
               else
                  call msg_fatal ("I/O error occurred while reading raw event file.")
               end if
               if (mcs%recalculate) then
                  call mc_recalculate (mcs, current_event)
               end if
            end if
            if (.not.mcs%read_events_raw) then
               call mc_generate (mcs, &
                    &            probability=eff(ch), &
                    &            weight=weight, channel=ch)
               if (mcs%write_events_raw) then
                  if (proc_id == WHIZARD_ROOT) &
                       & call write(uraw, current_event, fmt=RAW_FMT, &
                       &            probability=eff(ch), &
                       &            weight=weight, channel=ch)
               end if
            end if
            if (rho_in_allocated) then
               call destroy_rho_in (current_event)
               rho_in_allocated = .false.
            end if
            effsum(ch) = effsum(ch) + eff(ch)
            call fill_histogram (hh(ch), eff(ch))
            call fill_histogram (h, eff(ch))
            call apply (whizard_analysis(mc_process_index), &
                 &      current_event, weight/mcs%integral)
         end do
         if (proc_id == WHIZARD_ROOT) then
            call write_histogram (hh(ch), uwgt, over=.true.)
            write(uwgt,*) "!"
            if (int(n_calls * mcs%g%weights(ch)) > 0) then
               efftmp = effsum(ch) / int(n_calls * mcs%g%weights(ch))
            else
               efftmp = 0
            end if
            write(uwgt,*) "! Efficiency = ", efftmp
         end if
      end do
    end subroutine generate_weighted_single
    
  end subroutine generate_events_weighted

  subroutine whizard_end
    if (event_files_open)  call whizard_events_end
    deallocate (process_probability)
    if (mcs_exists) then
       call destroy (mcs)
       deallocate (mcs)
       call destroy (current_event)
       mcs_exists = .false.
    end if
    if (whizard_input_exists) then
       call destroy (whizard_input)
       whizard_input_exists = .false.
    end if
    call msg_summary
    call msg_message (" WHIZARD run finished.")
    call msg_count_clear
!    print *, "! WHIZARD run finished"
    if (mpi_initialized) then
       call mpi90_finalize
       mpi_initialized = .false.
    end if
  end subroutine whizard_end

end module whizard
